/* Gamma Data Compression (LZARI)

 Compression.C - Compression Function File

void StartModel(void);
void UpdateModel(int);
void StartDecode(void);
void Initialize(void);

int BinarySearchSym(unsigned int);
int BinarySearchPos(unsigned int);
int GetBit(void);
int DecodeChar(void);
int DecodePosition(void);
int LZARIDecompress(char *, char *);

*/

//////////////////////////////////////////////////////////////////////

void StartModel(void);
void UpdateModel(int);
void StartDecode(void);
void Initialize(void);

int BinarySearchSym(unsigned int);
int BinarySearchPos(unsigned int);
int GetBit(void);
int DecodeChar(void);
int DecodePosition(void);
int Compress(char *, char *);

//////////////////////////////////////////////////////////////////////

FILE *CompressSource, *CompressTarget;

/* Size Of Ring Buffer */
#define SizeOfRingBuffer 4096

/* Upper Limit For MatchLength */
#define MatchLengthLimit 60

/* Encode String Into Position And Length
    If MatchLength Is Greater Than This */
#define THRESHOLD        2

/* Ring Buffer Of Size SizeOfRingBuffer, With Extra MatchLengthLimit-1
    Bytes To Facilitate String Comparison */
unsigned char  TextBuffer[SizeOfRingBuffer + MatchLengthLimit - 1];

/* Used And Set By GetBit() Procedure */
unsigned int GetBitBuffer = 0, GetBitMask = 0;

///////////////////////////////////////////////////////////////////////
/********** Arithmetic Compression **********/

#define M   15

/* Q1 ( = 2 To The M) Must Be Sufficiently Large, But Not So
    Large As The Unsigned Long 4 * Q1 * (Q1 - 1) OverFlows.  */

#define Q1  (1UL << M)
#define Q2  (2 * Q1)
#define Q3  (3 * Q1)
#define Q4  (4 * Q1)
#define MAX_CUM (Q1 - 1)

/* Character Code = 0, 1, ..., N_CHAR - 1 */
#define N_CHAR  (256 - THRESHOLD + MatchLengthLimit)

unsigned long int Low = 0, High = Q4, Value = 0;

int Character2Symbol[N_CHAR], Symbol2Character[N_CHAR + 1];

    /* Frequency For Symbols */
unsigned int SymbolsFrequency[N_CHAR + 1];

    /* Cumulative Frequency For Symbols */
unsigned int CumulativeSymbols[N_CHAR + 1];

    /* Cumulative Frequency For Positions */
unsigned int CumulativePosition[SizeOfRingBuffer + 1];

//////////////////////////////////////////////////////////////////////

int GetBit(void)  /* Get one bit (0 or 1) */
{
    if ((GetBitMask >>= 1) == 0)
    {
        GetBitBuffer = getc(CompressSource);
        GetBitMask = 128;
    }
    return( (GetBitBuffer & GetBitMask) != 0);
}

///////////////////////////////////////////////////////////////////////
/********** Arithmetic Compression **********/

void StartModel(void)  /* Initialize model */
{
    int ch, sym, i;

    CumulativeSymbols[N_CHAR] = 0;
    for (sym = N_CHAR; sym >= 1; sym--)
    {
        ch = sym - 1;
        Character2Symbol[ch] = sym;
        Symbol2Character[sym] = ch;
        SymbolsFrequency[sym] = 1;
        CumulativeSymbols[sym - 1] = CumulativeSymbols[sym] + SymbolsFrequency[sym];
    }

    SymbolsFrequency[0] = 0;  /* sentinel (!= SymbolsFrequency[1]) */
    CumulativePosition[SizeOfRingBuffer] = 0;

    for (i = SizeOfRingBuffer; i >= 1; i--)
        CumulativePosition[i - 1] = CumulativePosition[i] + 10000 / (i + 200);

    /* Empirical Distribution Function (Quite Tentative) */
    /* Please Devise A Better Mechanism! */
}

///////////////////////////////////////////////////////////////////////

void UpdateModel(int sym)
{
    int i, c, ch_i, ch_sym;

    if (CumulativeSymbols[0] >= MAX_CUM)
    {
        c = 0;
        for (i = N_CHAR; i > 0; i--)
        {
            CumulativeSymbols[i] = c;
            c += (SymbolsFrequency[i] = (SymbolsFrequency[i] + 1) >> 1);
        }
        CumulativeSymbols[0] = c;
    }

    for (i = sym; SymbolsFrequency[i] == SymbolsFrequency[i - 1]; i--);

    if (i < sym)
    {
        ch_i = Symbol2Character[i];
        ch_sym = Symbol2Character[sym];
        Symbol2Character[i] = ch_sym;
        Symbol2Character[sym] = ch_i;
        Character2Symbol[ch_i] = sym;
        Character2Symbol[ch_sym] = i;
    }

    SymbolsFrequency[i]++;

    while (--i >= 0)
        CumulativeSymbols[i]++;
}

///////////////////////////////////////////////////////////////////////
/* 1      if x >= CumulativeSymbols[1],
N_CHAR if CumulativeSymbols[N_CHAR] > x,
i such that CumulativeSymbols[i - 1] > x >= CumulativeSymbols[i] otherwise */

int BinarySearchSym(unsigned int x)
{
    int i, j, k;

    i = 1;
    j = N_CHAR;
    while (i < j)
    {
        k = (i + j) / 2;
        if (CumulativeSymbols[k] > x)
            i = k + 1;
                else j = k;
    }
    return i;
}

///////////////////////////////////////////////////////////////////////
/* 0 if x >= CumulativePosition[1],
SizeOfRingBuffer - 1 if CumulativePosition[SizeOfRingBuffer] > x,
i such that CumulativePosition[i] > x >= CumulativePosition[i + 1] otherwise */

int BinarySearchPos(unsigned int x)
{
    int i, j, k;

    i = 1;
    j = SizeOfRingBuffer;
    while (i < j)
    {
        k = (i + j) / 2;
        if (CumulativePosition[k] > x)
            i = k + 1;
                else j = k;
    }
    return i - 1;
}

///////////////////////////////////////////////////////////////////////

void StartDecode(void)
{
    int i;
    for (i = 0; i < M + 2; i++)
        Value = 2 * Value + GetBit();
}

///////////////////////////////////////////////////////////////////////

int DecodeChar(void)
{
    int  sym, ch;
    unsigned long int  range;

    range = High - Low;
    sym = BinarySearchSym( (unsigned int) (((Value - Low + 1) * CumulativeSymbols[0] - 1) / range));
    High = Low + (range * CumulativeSymbols[sym - 1]) / CumulativeSymbols[0];
    Low +=       (range * CumulativeSymbols[sym    ]) / CumulativeSymbols[0];

    for (;;)
    {
        if (Low >= Q2)
        {
            Value -= Q2;
            Low -= Q2;
            High -= Q2;
        }
        else if (Low >= Q1 && High <= Q3)
        {
            Value -= Q1;
            Low -= Q1;
            High -= Q1;
        }
        else if (High > Q2) break;

        Low += Low;
        High += High;
        Value = 2 * Value + GetBit();
    }

    ch = Symbol2Character[sym];
    UpdateModel(sym);
    return ch;
}

///////////////////////////////////////////////////////////////////////

int DecodePosition(void)
{
    int position;
    unsigned long int  range;

    range = High - Low;
    position = BinarySearchPos((unsigned int) (((Value - Low + 1) * CumulativePosition[0] - 1) / range));
    High = Low + (range * CumulativePosition[position    ]) / CumulativePosition[0];
    Low +=       (range * CumulativePosition[position + 1]) / CumulativePosition[0];
    for ( ; ; )
    {
        if (Low >= Q2)
        {
            Value -= Q2;
            Low -= Q2;
            High -= Q2;
        }
        else if (Low >= Q1 && High <= Q3)
        {
            Value -= Q1;
            Low -= Q1;
            High -= Q1;
        }
        else if (High > Q2) break;

        Low += Low;
        High += High;
        Value = 2 * Value + GetBit();
    }
    return position;
}

///////////////////////////////////////////////////////////////////////

int LZARIDecompress(char *SourceFileName, char *TargetFileName)
{
    Initialize();

    int  i=0, j=0, k=0, r=0, c=0;


    unsigned long int PrintCount = 0, TextSize = 0;
    
    float StatusValue = 0;

    DWORD SourceFileSize = 0;
    long double Current_Size = 0;
    char StatusTemp[1024];

    if((CompressSource = fopen(SourceFileName,"rb")) == NULL)
    {
        if(CompressSource)fclose(CompressSource);
        ReportError("Compression Error: Could Read Open Source.");
        return 0;
    }

    if((CompressTarget = fopen(TargetFileName,"wb")) == NULL)
    {
        if(CompressTarget)fclose(CompressTarget);
        ReportError("Compression Error: Could Not Write Target.");
        return 0;
    }

    fseek(CompressSource, 0L, SEEK_END);
    SourceFileSize = ftell(CompressSource);
    fseek(CompressSource, 0L, SEEK_SET);

    /* read size of text */
    if(fread(&TextSize, sizeof TextSize, 1, CompressSource) < 1)
    {
        ReportError("Compression Error: Cannot Read Past Executable End.");
        return 0;
    }

    sprintf(StatusTemp,"Decompressing %lu KB to %lu KB \r\n", SourceFileSize/1024, TextSize/1024);
    SetText(StatusText_hWnd,StatusTemp);

    if(TextSize < 1)
    {
        ReportError("Compression Error: No Data Reported.");
        return 0;
    }

    SourceFileSize = TextSize;

    StartDecode();
    StartModel();

    for(i = 0; i < SizeOfRingBuffer - MatchLengthLimit; i++) TextBuffer[i] = ' ';

    r = SizeOfRingBuffer - MatchLengthLimit;

    for(Current_Size = 0; Current_Size < TextSize; )
    {
        c = DecodeChar();
        if (c < 256)
        {
            putc(c, CompressTarget);
            TextBuffer[r++] = c;
            r &= (SizeOfRingBuffer - 1);
            Current_Size++;
        }
        else{
            i = (r - DecodePosition() - 1) & (SizeOfRingBuffer - 1);
            j = c - 255 + THRESHOLD;
            for (k = 0; k < j; k++) {
                c = TextBuffer[(i + k) & (SizeOfRingBuffer - 1)];
                putc(c, CompressTarget);
                TextBuffer[r++] = c;
                r &= (SizeOfRingBuffer - 1);
                Current_Size++;
            }
        }

        if (Current_Size > PrintCount)
        {
            if(CancelFlag)
            {
                if(CompressSource)fclose(CompressSource);
                if(CompressTarget)fclose(CompressTarget);
                return 0;
            }

            StatusValue = (float)((Current_Size / SourceFileSize) * 100);
            SendMessage(StatusBar_hWnd, PBM_SETPOS,(DWORD)StatusValue,0);
            PrintCount += 1024;
        }
    }

    /* Update Status One Last Time */
    StatusValue = (float)((Current_Size / SourceFileSize) * 100);
    SendMessage(StatusBar_hWnd, PBM_SETPOS,(DWORD)StatusValue,0);

    if(CompressSource)fclose(CompressSource);
    if(CompressTarget)fclose(CompressTarget);

    return 1;
}

///////////////////////////////////////////////////////////////////////

void Initialize(void)
{

    if(CompressSource)fclose(CompressSource);
    if(CompressTarget)fclose(CompressTarget);

    CompressSource = 0;
    CompressTarget = 0;

    GetBitBuffer = 0;
    GetBitMask = 0;

    Low = 0;
    High = Q4;
    Value = 0;

    DWORD LocalClearLoop = 0;

    while(LocalClearLoop != SizeOfRingBuffer + MatchLengthLimit - 1)
    {
        TextBuffer[LocalClearLoop] = '\0';
        LocalClearLoop++;
    }LocalClearLoop = 0;

    while(LocalClearLoop != N_CHAR)
    {
        Character2Symbol[LocalClearLoop] = 0;
        LocalClearLoop++;
    }LocalClearLoop = 0;

    while(LocalClearLoop != N_CHAR + 1)
    {
        Symbol2Character[LocalClearLoop] = 0;
        LocalClearLoop++;
    }LocalClearLoop = 0;

    while(LocalClearLoop != N_CHAR + 1)
    {
        SymbolsFrequency[LocalClearLoop] = 0;
        LocalClearLoop++;
    }LocalClearLoop = 0;

    while(LocalClearLoop != N_CHAR + 1)
    {
        CumulativeSymbols[LocalClearLoop] = 0;
        LocalClearLoop++;
    }LocalClearLoop = 0;

    while(LocalClearLoop != SizeOfRingBuffer + 1)
    {
        CumulativePosition[LocalClearLoop] = 0;
        LocalClearLoop++;
    }
}

///////////////////////////////////////////////////////////////////////

